/* sgfile.c
 *
 * ScanGroup file code (also handles restriction entries)
 */

/* trim the list */
#include "EXTERN.h"
#include "common.h"
#include "intrp.h"
#include "search.h"
#include "sgroup.h"
#include "sgdata.h"
#include "sgmisc.h"
#include "term.h"
#include "util.h"
#include "INTERN.h"
#include "sgfile.h"

/* complex expressions taken from ngsrch.c */

static COMPEX sgcompex;

static sg_file_initialized	INIT(FALSE);

/* initializes sg_file stuff */
void
sg_file_init()
{
    if (sg_file_initialized)
	return;			/* only necessary once */
    init_compex(&sgcompex);
    sg_file_initialized = TRUE;
}

bool sg_match_err	INIT(FALSE);

/* Wildmat function:
 *      Based on Rich $alz's wildmat, reduced to the simple case of *
 *      and text.  The complete version can be found in Rich's INN,
 *      among other places; write to <rsalz@uunet.uu.net>
 */

/* an improbable number */
#define ABORT			-42


/*
**  Match text and p, return TRUE, FALSE, or ABORT.
*/
static int
sg_DoMatch(text, p)
register char	*text;
register char	*p;
{
    register int	matched;

    for ( ; *p; text++, p++) {
	if (*p == '*') {
	    while (*++p == '*')
		/* Consecutive stars act just like one. */
		continue;
	    if (*p == '\0')
		/* Trailing star matches everything. */
		return TRUE;
	    while (*text)
		if ((matched = sg_DoMatch(text++, p)) != FALSE)
		    return matched;
	    return ABORT;
	}
	if (*text != *p) {
	    if (*text == '\0')
		return ABORT;
	    return FALSE;
	}
    }
    return *text == '\0';
}

void
sg_use_pattern(pattern)
char *pattern;
{
    register long i;
    long e;		/* misc. entry */
    char ng_pattern[128];
    register char *s = pattern, *d = ng_pattern;

    sg_match_err = FALSE;		/* no errors (yet :-) */
    if (!s || !*s) {
	printf("\nScanGroup match: empty regular expression\n") FLUSH;
	return;
    }
    if (*pattern == '!') {	/* exclude these matches */
	if (!*(pattern+1)) {
	    printf("\nScanGroup match: empty expression after '!'\n") FLUSH;
	    return;
	}

	/* note: skip zero'th entry */
	for (i=1;i<sg_num_ents;i++) {
	    if (sg_ents[i].type == 1)		/* newsgroup type */
		if (sg_DoMatch(sg_groups[sg_ents[i].groupnum].name,pattern+1)
		    == TRUE)
			sg_ents[i].type = -1;	/* make entry invalid */
	}
    } else {	/* include these matches */
	for (i=0;i<sg_num_groups;i++) {
	    if (sg_DoMatch(sg_groups[i].name,pattern) == TRUE) {
		int j;
		for (j=1;j<sg_num_ents;j++)
		    if (sg_ents[j].type == 1)	/* newsgroup type */
			if (sg_ents[j].groupnum == i)	/* same group */
			    break;	/* out of for loop */
		if (j>=sg_num_ents) {	/* was not there before */
		    e = sg_add_ent(1,sg_groups[i].name);
		    sg_ents[e].groupnum = i;
		}
	    }
	}
    }
}

void
sg_use_restrict(restrict,title)
char *restrict;
char *title;
{
    (void)sg_new_context();
    sg_contexts[sg_num_contexts-1].title = savestr(title);

    sg_use_group_line(restrict);
    sg_delete_invalid_ents();			/* in case some were deleted */
}

/* returns TRUE on success, FALSE otherwise */
bool
sg_use_file(fname,title)
char *fname;		/* file name */
char *title;		/* for top bar */
{
    static char lbuf[1024];
    FILE *fp;
    char *s,*p;
    int i;

    s = fname;
    if (*s == ':') {	/* relative to last file's directory */
	if (sg_num_contexts==0) {
	    /* really shouldn't happen */
	    printf("\nsg_use_file: top level filename error\n") FLUSH;
	    return(FALSE);
	}
	p = sg_contexts[sg_num_contexts-1].filename;
	for (i=sg_num_contexts-1;i>=0;i--)
	    if ((p=sg_contexts[i].filename))	/* EQUALS */
		break;
	if (i<0) {	/* no file found */
	    printf("\nsg_use_file internal error: current not found\n") FLUSH;
	    /* consider doing an assertion? something is really wrong */
	    return(FALSE);
	}
	strcpy(lbuf,p);
	p = rindex(lbuf,'/');
	if (!p) {
	    printf("\nsg_use_file: current has no slash\n") FLUSH;
	    return(FALSE);
	}
	p++;					/* point to after slash */
	strcpy(p,s+1);
	s = lbuf;
    }
/*    printf("\nAttempting to open file %s\n",s) FLUSH; /* */
    fp = fopen(filexp(s),"r");
    if (!fp)
	return(FALSE);		/* unsuccessful */
    (void)sg_new_context();
    sg_contexts[sg_num_contexts-1].title = savestr(title);
    sg_contexts[sg_num_contexts-1].filename = savestr(s);

/* Later considerations:
 * 1. Long lines
 * 2. Backslash continuations
 */
    while ((s = fgets(lbuf,1020,fp))) {		/* EQUALS */
	if (!s)		/* end of file */
	    break;
	sg_do_line(s);
    }
    fclose(fp);
    sg_delete_invalid_ents();			/* in case some were deleted */
    return(TRUE);
}

/* if non-NULL, is the description (printing name) of the entry */
static char *sg_line_desc;

void
sg_do_line(line)
char *line;
{
    char *s;
    char *p;
    long e;		/* entry index */

    s = line;

    if (s[strlen(s)-1] == '\n')
	s[strlen(s)-1] = '\0';		/* delete newline */
    while ((*s == ' ') || (*s == '\t'))
	s++;		/* delete whitespace */

    if (sg_line_desc) {
	free(sg_line_desc);
	sg_line_desc = (char*)NULL;
    }
    if (*s == '"') {	/* description name */
	/* try handling backslashes later?  Perhaps other special cases? */
	s++;
	for (p=s;*p && *p != '"';p++)
	    ;	/* EMPTY */
	if (!*p) {
	    printf("ScanGroup: unfinished quote.  Line was:\n%s\n",
		   line) FLUSH;
	    return;
	}
	*p = '\0';
	sg_line_desc = savestr(s);
	*p = '"';
	++p;
	s = p;
    }
    while ((*s == ' ') || (*s == '\t'))
	s++;		/* delete whitespace */
    switch (*s) {
	case '#':	/* comment */
	    break;
	case ':':	/* relative to sg_root */
	case '~':
	case '/':
	    /* description defaults to name */
	    if (sg_line_desc)
		e = sg_add_ent(2,sg_line_desc);
	    else
		e = sg_add_ent(2,s);
	    sg_ents[e].restrict = savestr(s);
	    break;
	default:
	    /* if there is a description, this must be a restriction list */
	    if (sg_line_desc) {
		e = sg_add_ent(3,sg_line_desc);
		sg_ents[e].restrict = savestr(s);
		break;
	    }
	    /* one or more newsgroups instead */
	    sg_use_group_line(s);
	    break;
    }
}

/* interprets a line of newsgroups, adding or subtracting each pattern */
/* Newsgroup patterns are separated by spaces */
void
sg_use_group_line(line)
char *line;
{
    char *s,*p;
    char ch;
    int i,j;

/*    printf("sg_use_group_line(%s) called\n",line); /* */
    s = line;
    if (!s || !*s)
	return;
    /* newsgroup patterns will be separated by space(s) */
    while (*s) {
	for (p=s;*p && (*p != ' ');p++)
	    ; /* EMPTY */
	ch = *p;
	*p = '\0';
	sg_use_pattern(s);
	*p = ch;
	s = p;
	while (*s == ' ')
	    s++;
    }
}

/* edits the current file */
void
sg_edit_file()
{
    char *fname;
    char *s;
    int i;
    char lbuf[1024];	/* hold the filename */
    char wherebuf[1024];	/* where were we? */

    for (i=sg_num_contexts-1;i>=0;i--)
	if ((fname=sg_contexts[i].filename))	/* EQUALS */
	    break;
    if (i<0) {	/* no file found */
	printf("\nsg_edit_file internal error: current not found\n") FLUSH;
	/* consider doing an assertion? something is really wrong */
	return;
    }
    if (!fname || !*fname)
	return;		/* empty, do nothing (error later?) */
    strcpy(lbuf,filexp(fname));
    sprintf(cmd_buf,"%s %s",
	filexp(getval("VISUAL",getval("EDITOR",defeditor))),lbuf);
    printf("\nEditing ScanGroup file %s:\n%s\n",fname,cmd_buf) FLUSH;
    resetty();			/* make sure tty is friendly */
    getwd(wherebuf);	/* find out where we are */
    s = rindex(lbuf,'/');
    if (!s)
	return;		/* error, but not critical */
    if (s == lbuf)	/* wow, file in root directory */
	s++;		/* will need the slash */
    *s = '\0';	/* cut off after directory name */
    if (chdir(lbuf)) {
	printf(nocd,lbuf) FLUSH;
	return;
    }
    doshell(sh,cmd_buf);/* invoke the shell */
    if (chdir(wherebuf)) {
	printf(nocd,wherebuf) FLUSH;
	return;
    }
    noecho();			/* and make terminal */
    crmode();			/*   unfriendly again */
}
